/*
Copyright 2004 John Tsiombikas <nuclear@siggraph.org>

This file is part of libgba, a library for GameBoy Advance development.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "libgba_config.h"

#include <stdlib.h>
#include "gfx.h"

#define FRAME_SEL_BIT	0x10

#define H_BLANK_OAM		0x20
#define OBJ_MAP_2D		0x0
#define OBJ_MAP_1D		0x40
#define FORCE_BLANK		0x80
#define BG0_ENABLE		0x100
#define BG1_ENABLE		0x200
#define BG2_ENABLE		0x400
#define BG3_ENABLE		0x800

static volatile unsigned short *reg_disp_ctl = (void*)0x4000000;
static volatile unsigned short *reg_vcount = (void*)0x4000006;

static int xres, yres;
static int sizeof_pixel;
static int page_flipping;
static int allocated_back_buffer;

static struct pixel_buffer fbuf, bbuf;
struct pixel_buffer *front_buffer = &fbuf;
struct pixel_buffer *back_buffer = &bbuf;

#define show_page(n)	((n) ? (*reg_disp_ctl |= FRAME_SEL_BIT) : (*reg_disp_ctl &= ~FRAME_SEL_BIT))
#define swap_page()		(*reg_disp_ctl ^= FRAME_SEL_BIT)

int set_video_mode(int mode, int double_buffering) {
	if(mode < 3 || mode > 5) return -1;

	/* mode 5: 160x128, otherwise: 240x160 */
	xres = (mode == 5) ? 160 : 240;
	yres = (mode == 5) ? 128 : 160;

	sizeof_pixel = (mode == 4) ? 1 : 2;

	*reg_disp_ctl = mode | BG2_ENABLE;

	show_page(0);

	if(allocated_back_buffer) {
		free(back_buffer->pixels);
	}

	front_buffer->pixels = (void*)0x6000000;
	front_buffer->x = back_buffer->x = xres;
	front_buffer->y = back_buffer->y = yres;
	front_buffer->bpp = back_buffer->bpp = sizeof_pixel * 8;
		
	if(mode > 3) {
		page_flipping = 1;
		back_buffer->pixels = (void*)0x600a000;
	} else {
		page_flipping = 0;
		if(double_buffering) {
			back_buffer->pixels = malloc(xres * yres * sizeof_pixel);
		}
	}
}

void flip(void) {
	static void *tmp;
	
	if(page_flipping) {
		swap_page();
		tmp = front_buffer->pixels;
		front_buffer->pixels = back_buffer->pixels;
		back_buffer->pixels = tmp;
	} else {
		dma_copy32(3, front_buffer->pixels, back_buffer->pixels, (xres * yres) >> 1);
	}
}

/* ------- pixel buffer operations ------- */

struct pixel_buffer *create_pixel_buffer(int x, int y, int bpp) {
	struct pixel_buffer *pbuf = malloc(sizeof(struct pixel_buffer));
	if(pbuf) {
		pbuf->pixels = malloc(x * y * bpp / 8);
		pbuf->x = x;
		pbuf->y = y;
		pbuf->bpp = bpp;
	}
	return pbuf;
}

void destroy_pixel_buffer(struct pixel_buffer *pbuf) {
	free(pbuf->pixels);
	free(pbuf);
}

void clear_buffer(struct pixel_buffer *pbuf, unsigned short color) {
	int sz = pbuf->x * pbuf->y;

	if(pbuf->bpp == 8) {
		color |= color << 8;
		sz >>= 1;
	}
		
	dma_fill16(3, pbuf->pixels, color, sz);
}

void copy_buffer(const struct pixel_buffer *src, struct pixel_buffer *dst) {
	int words;
	
	if(src->x != dst->x || src->y != dst->y || src->bpp != dst->bpp) return;

	words = (src->x * src->y) >> (src->bpp == 16 ? 1 : 2);
	dma_copy32(3, dst->pixels, src->pixels, words);
}
